==> editex-0.2.0.tar.gz: lib/src/utils.dart <==
// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

library editex.utils;

import 'package:eqlib/latex.dart';
import 'package:editex/interface.dart';

const _cursorLaTeX = r'\phantom{c}';
const _placeholderLaTeX = r'\phantom{e}';
const _autocompleteLaTeX = r'\phantom{a}';

class StackItem {
  final LaTeXTemplate latexTemplate;
  final String parsableTemplate;
  final arguments = new List<List<String>>();

  StackItem(this.latexTemplate, this.parsableTemplate) {
    arguments.add(new List<String>());
  }
}

String renderLaTeXTemplate(LaTeXTemplate template, List<String> arguments) {
  final buffer = new StringBuffer();

  for (final token in template.tokens) {
    if (token.text.isNotEmpty) {
      buffer.write(token.text);
    } else {
      final parentheses = token.parenthesesPriority != -1 ||
          token.paramLeftBaseline ||
          token.paramRightBaseline;

      if (parentheses) {
        buffer.write(r'\left(');
      }

      final argTex = arguments[token.paramIndex];
      buffer.write(argTex);
      if (argTex.isEmpty || argTex == _cursorLaTeX) {
        buffer.write(_placeholderLaTeX);
      }

      if (parentheses) {
        buffer.write(r'\right)');
      }
    }
  }

  final tex = buffer.toString();

  // Render as block or else spacing will be affected when the cursor is
  // inserted after this (since the cursor is wrapped in a block with the
  // previous element to not affect height).
  if (new RegExp(r'[\^_]').matchAsPrefix(tex) == null) {
    return '{$tex}';
  } else {
    return '$tex';
  }
}

String renderParsableTemplate(String template, List<String> arguments) {
  return template.replaceAllMapped(new RegExp(r'\$([0-9]+)'), (match) {
    return arguments[int.parse(match.group(1))];
  });
}

/// Generate string expression from editor data. Uses LaTeX templates if [latex]
/// is true. Else uses the parsing template. Only inserts cursors in LaTeX
/// expressions.
String generateString(bool latex, List data, Map<int, EdiTeXTemplate> templates,
    [List<int> cursors = const [],
    String selectionColor = 'red',
    int autocompleteAt = -1]) {
  var passedCursors = 0;
  final stack = new List<StackItem>();
  stack.add(new StackItem(parseLaTeXTemplate(r'${0}', null), r'$0'));

  for (var i = 0; i < data.length; i++) {
    // Add cursor.
    if (latex && cursors.contains(i)) {
      passedCursors++;

      // Group cursor together with previous part in this argument. This ensures
      // that sub/superscript stays on the same height regardless of the cursor.
      final argList = stack.last.arguments.last;
      if (argList.isNotEmpty &&
          new RegExp(r'[\^_]').matchAsPrefix(argList.last) == null) {
        argList.insert(argList.length - 1, '{');
        argList.add(_cursorLaTeX);
        argList.add('}');
      } else {
        argList.add(_cursorLaTeX);
      }

      // Add selection color.
      if (passedCursors % 2 == 1 && cursors.length - passedCursors > 0) {
        argList.add('\\color{$selectionColor}{');
      } else if (passedCursors % 2 == 0) {
        argList.add('}');
      }
    }

    // Add autocomplete.
    if (latex && autocompleteAt == i) {
      stack.last.arguments.last.add(_autocompleteLaTeX);
    }

    // Add element.
    final elm = data[i];
    if (elm is String) {
      // Append text to last argument in last stack element.
      if (latex) {
        final tex = elm == r'\' ? r'\backslash' : '\\text{$elm}';
        stack.last.arguments.last.add(tex);
      } else {
        stack.last.arguments.last.add(elm);
      }
    } else if (elm is int) {
      // Number can indicate a template ID (>=0)
      // or template argument boundary (<0).
      if (elm >= 0) {
        // Add template to stack.
        final t = templates[elm];
        stack.add(new StackItem(t.template, t.parsableTemplate));
      } else {
        // Go to next argument.
        stack.last.arguments.add(new List<String>());
      }
    }

    // Render template if all arguments are in.
    if (stack.last.arguments.length > stack.last.latexTemplate.parameterCount) {
      final arguments = stack.last.arguments.map((arg) => arg.join()).toList();
      final str = latex
          ? renderLaTeXTemplate(stack.last.latexTemplate, arguments)
          : renderParsableTemplate(stack.last.parsableTemplate, arguments);

      stack.removeLast();
      stack.last.arguments.last.add(str);
    }
  }

  // Terminal cursor.
  if (latex && cursors.contains(data.length)) {
    passedCursors++;
    stack.last.arguments.last.add(_cursorLaTeX);
    if (passedCursors % 2 == 0) {
      stack.last.arguments.last.add('}');
    }
  }

  return stack.single.arguments.single.join();
}

String getKeyword(List data, int cursor) {
  final buffer = new List<String>();
  for (var i = cursor - 1; i >= 0; i--) {
    if (data[i] is String) {
      buffer.insert(0, data[i]);
      if (data[i] == r'\') {
        return buffer.join();
      }
    } else {
      break;
    }
  }
  return buffer.join();
}

/// Find first index after given template.
int findTemplateEnd(
    List data, int startIndex, Map<int, EdiTeXTemplate> templates,
    [int stopAfterNParams]) {
  var i = startIndex;

  final templateId = data[i++];
  assert(templateId >= 0);

  final template = templates[templateId];
  var passedArgs = 0;

  final paramCount = stopAfterNParams ?? template.parameterCount;
  while (passedArgs < paramCount) {
    final removed = data[i++];
    if (removed is int) {
      if (removed >= 0) {
        final removedTemplate = templates[removed];
        passedArgs -= removedTemplate.parameterCount;
      } else {
        passedArgs++;
      }
    }
  }

  return i;
}

/// Iterate backward to the template ID from the given argument boundary.
/// Returns the index of the template ID.
int findTemplateStart(
    List data, int boundaryIndex, Map<int, EdiTeXTemplate> templates) {
  final int argumentCount = -data[boundaryIndex];
  var argumentsLeft = argumentCount;
  var i = boundaryIndex;
  while (i >= 0) {
    final value = data[--i];
    if (value is int) {
      if (value >= 0) {
        final t = templates[value];
        if (argumentsLeft == 1 && t.parameterCount >= argumentCount) {
          break;
        } else {
          argumentsLeft += templates[value].parameterCount;
        }
      } else {
        argumentsLeft--;
      }
    }
  }

  return i;
}

/// Returns new cursor index.
int insertTemplate(List data, int index, EdiTeXTemplate template) {
  data.insert(index++, template.id);

  // Insert placeholders.
  data.insertAll(
      index, new List<int>.generate(template.parameterCount, (i) => -1 - i));

  return index;
}

/// Returns new cursor index.
int removeElement(
    List data, int index, Map<int, EdiTeXTemplate> templates, int direction) {
  // If index is an integer, we need to do some special stuff.
  if (data[index] is int) {
    // If this is an argument boundary, first iterate all the way back to the
    // template ID.
    final templateStart =
        data[index] < 0 ? findTemplateStart(data, index, templates) : index;
    final templateEnd = findTemplateEnd(data, templateStart, templates);

    // Figure out if this template was deleted from the outside.
    final deletedFromOutside = (index == templateStart && direction == 1) ||
        (index == templateEnd - 1 && direction == -1);

    // Keep first argument if the template only has one parameter or is a
    // fraction. This is not the case if the given element is part of a template
    // and was deleted from the outside.
    final template = templates[data[templateStart]];
    if (!deletedFromOutside &&
        (template.parameterCount == 1 || template.type == EdiTeXTT.frac)) {
      final firstBoundaryEndIdx =
          findTemplateEnd(data, templateStart, templates, 1);
      data.removeRange(firstBoundaryEndIdx - 1, templateEnd);
      data.removeAt(templateStart);

      // If deletion target template start, retain cursor position.
      if (index == templateStart) {
        return index;
      } else {
        return firstBoundaryEndIdx - 2;
      }
    } else {
      data.removeRange(templateStart, templateEnd);
      return templateStart;
    }
  } else {
    data.removeAt(index);
    return index;
  }
}

/// Move selection cursor.
int moveSelectionCursor(List data, int currentIndex, int direction,
    Map<int, EdiTeXTemplate> templates) {
  assert(direction == 1 || direction == -1);
  final newIndex = currentIndex + direction;

  if (newIndex < 0) {
    return 0;
  } else if (newIndex >= data.length) {
    return data.length;
  } else {
    if (direction == 1) {
      final valueBefore = data[newIndex - 1];
      if (valueBefore is int && valueBefore >= 0) {
        // We passed a template ID. Pass the entire template.
        return findTemplateEnd(data, newIndex - 1, templates);
      } else if (valueBefore is int && valueBefore < 0) {
        // We passed an argument boundary, undo this.
        return currentIndex;
      } else {
        return newIndex;
      }
    } else {
      final valueAt = data[newIndex];
      if (valueAt is int && valueAt >= 0) {
        // This is either a 0 argument template or the argument boundary.
        if (templates[valueAt].parameterCount == 0) {
          return newIndex;
        } else {
          // Undo move.
          return currentIndex;
        }
      } else if (valueAt is int && valueAt < 0) {
        // Iterate back to the template ID, then check if this boundary is the
        // last boundary of this template. If so, return template ID index. Else
        // return old selection cursor.
        final templateStart = findTemplateStart(data, newIndex, templates);
        final templateId = data[templateStart];
        if (-valueAt == templates[templateId].parameterCount) {
          return templateStart;
        } else {
          return currentIndex;
        }
      } else {
        return newIndex;
      }
    }
  }
}

/// Get all empty argument indices.
List<int> getEmptyArguments(List data, Map<int, EdiTeXTemplate> templates) {
  final stack = new List<int>();
  final ret = new List<int>();

  var gap = false;
  for (var i = 0; i < data.length; i++) {
    final v = data[i];
    if (v is int) {
      if (v >= 0) {
        stack.add(templates[v].parameterCount);
        gap = false;
      } else {
        if (!gap) {
          ret.add(i);
        }

        stack[stack.length - 1]--;
        if (stack.last == 0) {
          stack.removeLast();
          gap = true;
        } else {
          gap = false;
        }
      }
    } else {
      gap = true;
    }
  }

  return ret;
}

==> editex-0.2.1.tar.gz: lib/src/utils.dart <==
// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

library editex.utils;

import 'package:eqlib/latex.dart';
import 'package:editex/interface.dart';

const _cursorLaTeX = r'\phantom{c}';
const _placeholderLaTeX = r'\phantom{e}';
const _autocompleteLaTeX = r'\phantom{a}';

class StackItem {
  final LaTeXTemplate latexTemplate;
  final String parsableTemplate;
  final arguments = new List<List<String>>();

  StackItem(this.latexTemplate, this.parsableTemplate) {
    arguments.add(new List<String>());
  }
}

String renderLaTeXTemplate(LaTeXTemplate template, List<String> arguments) {
  final buffer = new StringBuffer();

  for (final token in template.tokens) {
    if (token.text.isNotEmpty) {
      buffer.write(token.text);
    } else {
      final parentheses = token.parenthesesPriority != -1 ||
          token.paramLeftBaseline ||
          token.paramRightBaseline;

      if (parentheses) {
        buffer.write(r'\left(');
      }

      final argTex = arguments[token.paramIndex];
      buffer.write(argTex);
      if (argTex.isEmpty || argTex == _cursorLaTeX) {
        buffer.write(_placeholderLaTeX);
      }

      if (parentheses) {
        buffer.write(r'\right)');
      }
    }
  }

  final tex = buffer.toString();

  // Render as block or else spacing will be affected when the cursor is
  // inserted after this (since the cursor is wrapped in a block with the
  // previous element to not affect height).
  if (new RegExp(r'[\^_]').matchAsPrefix(tex) == null) {
    return '{$tex}';
  } else {
    return '$tex';
  }
}

String renderParsableTemplate(String template, List<String> arguments) {
  return template.replaceAllMapped(new RegExp(r'\$([0-9]+)'), (match) {
    return arguments[int.parse(match.group(1))];
  });
}

/// Generate string expression from editor data. Uses LaTeX templates if [latex]
/// is true. Else uses the parsing template. Only inserts cursors in LaTeX
/// expressions.
String generateString(bool latex, List data, Map<int, EdiTeXTemplate> templates,
    [List<int> cursors = const [],
    String selectionColor = 'red',
    int autocompleteAt = -1]) {
  var passedCursors = 0;
  final stack = new List<StackItem>();
  stack.add(new StackItem(parseLaTeXTemplate(r'${0}', null), r'$0'));

  for (var i = 0; i < data.length; i++) {
    // Add cursor.
    if (latex && cursors.contains(i)) {
      passedCursors++;

      // Group cursor together with previous part in this argument. This ensures
      // that sub/superscript stays on the same height regardless of the cursor.
      final argList = stack.last.arguments.last;
      if (argList.isNotEmpty &&
          new RegExp(r'[\^_]').matchAsPrefix(argList.last) == null) {
        argList.insert(argList.length - 1, '{');
        argList.add(_cursorLaTeX);
        argList.add('}');
      } else {
        argList.add(_cursorLaTeX);
      }

      // Add selection color.
      if (passedCursors % 2 == 1 && cursors.length - passedCursors > 0) {
        argList.add('\\color{$selectionColor}{');
      } else if (passedCursors % 2 == 0) {
        argList.add('}');
      }
    }

    // Add autocomplete.
    if (latex && autocompleteAt == i) {
      stack.last.arguments.last.add(_autocompleteLaTeX);
    }

    // Add element.
    final elm = data[i];
    if (elm is String) {
      // Append text to last argument in last stack element.
      if (latex) {
        final tex = elm == r'\' ? r'\backslash' : '\\text{$elm}';
        stack.last.arguments.last.add(tex);
      } else {
        stack.last.arguments.last.add(elm);
      }
    } else if (elm is int) {
      // Number can indicate a template ID (>=0)
      // or template argument boundary (<0).
      if (elm >= 0) {
        // Add template to stack.
        final t = templates[elm];
        stack.add(new StackItem(t.template, t.parsableTemplate));
      } else {
        // Go to next argument.
        stack.last.arguments.add(new List<String>());
      }
    }

    // Render template if all arguments are in.
    if (stack.last.arguments.length > stack.last.latexTemplate.parameterCount) {
      final arguments = stack.last.arguments.map((arg) => arg.join()).toList();
      final str = latex
          ? renderLaTeXTemplate(stack.last.latexTemplate, arguments)
          : renderParsableTemplate(stack.last.parsableTemplate, arguments);

      stack.removeLast();
      stack.last.arguments.last.add(str);
    }
  }

  // Terminal cursor.
  if (latex && cursors.contains(data.length)) {
    passedCursors++;
    stack.last.arguments.last.add(_cursorLaTeX);
    if (passedCursors % 2 == 0) {
      stack.last.arguments.last.add('}');
    }
  }

  return stack.single.arguments.single.join();
}

String getKeyword(List data, int cursor) {
  final buffer = new List<String>();
  var pastDigits = false;
  for (var i = cursor - 1; i >= 0; i--) {
    if (data[i] is String) {
      if (new RegExp('[0-9]').hasMatch(data[i])) {
        if (pastDigits) {
          break;
        }
      } else {
        pastDigits = true;
      }

      buffer.insert(0, data[i]);
      if (data[i] == r'\') {
        return buffer.join();
      }
    } else {
      break;
    }
  }
  return buffer.join();
}

/// Find first index after given template.
int findTemplateEnd(
    List data, int startIndex, Map<int, EdiTeXTemplate> templates,
    [int stopAfterNParams]) {
  var i = startIndex;

  final templateId = data[i++];
  assert(templateId >= 0);

  final template = templates[templateId];
  var passedArgs = 0;

  final paramCount = stopAfterNParams ?? template.parameterCount;
  while (passedArgs < paramCount) {
    final removed = data[i++];
    if (removed is int) {
      if (removed >= 0) {
        final removedTemplate = templates[removed];
        passedArgs -= removedTemplate.parameterCount;
      } else {
        passedArgs++;
      }
    }
  }

  return i;
}

/// Iterate backward to the template ID from the given argument boundary.
/// Returns the index of the template ID.
int findTemplateStart(
    List data, int boundaryIndex, Map<int, EdiTeXTemplate> templates) {
  final int argumentCount = -data[boundaryIndex];
  var argumentsLeft = argumentCount;
  var i = boundaryIndex;
  while (i >= 0) {
    final value = data[--i];
    if (value is int) {
      if (value >= 0) {
        final t = templates[value];
        if (argumentsLeft == 1 && t.parameterCount >= argumentCount) {
          break;
        } else {
          argumentsLeft += templates[value].parameterCount;
        }
      } else {
        argumentsLeft--;
      }
    }
  }

  return i;
}

/// Returns new cursor index.
int insertTemplate(List data, int index, EdiTeXTemplate template) {
  data.insert(index++, template.id);

  // Insert placeholders.
  data.insertAll(
      index, new List<int>.generate(template.parameterCount, (i) => -1 - i));

  return index;
}

/// Returns new cursor index.
int removeElement(
    List data, int index, Map<int, EdiTeXTemplate> templates, int direction) {
  // If index is an integer, we need to do some special stuff.
  if (data[index] is int) {
    // If this is an argument boundary, first iterate all the way back to the
    // template ID.
    final templateStart =
        data[index] < 0 ? findTemplateStart(data, index, templates) : index;
    final templateEnd = findTemplateEnd(data, templateStart, templates);

    // Figure out if this template was deleted from the outside.
    final deletedFromOutside = (index == templateStart && direction == 1) ||
        (index == templateEnd - 1 && direction == -1);

    // Keep first argument if the template only has one parameter or is a
    // fraction. This is not the case if the given element is part of a template
    // and was deleted from the outside.
    final template = templates[data[templateStart]];
    if (!deletedFromOutside &&
        (template.parameterCount == 1 || template.type == EdiTeXTT.frac)) {
      final firstBoundaryEndIdx =
          findTemplateEnd(data, templateStart, templates, 1);
      data.removeRange(firstBoundaryEndIdx - 1, templateEnd);
      data.removeAt(templateStart);

      // If deletion target template start, retain cursor position.
      if (index == templateStart) {
        return index;
      } else {
        return firstBoundaryEndIdx - 2;
      }
    } else {
      data.removeRange(templateStart, templateEnd);
      return templateStart;
    }
  } else {
    data.removeAt(index);
    return index;
  }
}

/// Move selection cursor.
int moveSelectionCursor(List data, int currentIndex, int direction,
    Map<int, EdiTeXTemplate> templates) {
  assert(direction == 1 || direction == -1);
  final newIndex = currentIndex + direction;

  if (newIndex < 0) {
    return 0;
  } else if (newIndex >= data.length) {
    return data.length;
  } else {
    if (direction == 1) {
      final valueBefore = data[newIndex - 1];
      if (valueBefore is int && valueBefore >= 0) {
        // We passed a template ID. Pass the entire template.
        return findTemplateEnd(data, newIndex - 1, templates);
      } else if (valueBefore is int && valueBefore < 0) {
        // We passed an argument boundary, undo this.
        return currentIndex;
      } else {
        return newIndex;
      }
    } else {
      final valueAt = data[newIndex];
      if (valueAt is int && valueAt >= 0) {
        // This is either a 0 argument template or the argument boundary.
        if (templates[valueAt].parameterCount == 0) {
          return newIndex;
        } else {
          // Undo move.
          return currentIndex;
        }
      } else if (valueAt is int && valueAt < 0) {
        // Iterate back to the template ID, then check if this boundary is the
        // last boundary of this template. If so, return template ID index. Else
        // return old selection cursor.
        final templateStart = findTemplateStart(data, newIndex, templates);
        final templateId = data[templateStart];
        if (-valueAt == templates[templateId].parameterCount) {
          return templateStart;
        } else {
          return currentIndex;
        }
      } else {
        return newIndex;
      }
    }
  }
}

/// Get all empty argument indices.
List<int> getEmptyArguments(List data, Map<int, EdiTeXTemplate> templates) {
  final stack = new List<int>();
  final ret = new List<int>();

  var gap = false;
  for (var i = 0; i < data.length; i++) {
    final v = data[i];
    if (v is int) {
      if (v >= 0) {
        final parameterCount = templates[v].parameterCount;
        if (parameterCount > 0) {
          stack.add(parameterCount);
          gap = false;
        } else {
          gap = true;
        }
      } else {
        if (!gap) {
          ret.add(i);
        }

        stack[stack.length - 1]--;
        if (stack.last == 0) {
          stack.removeLast();
          gap = true;
        } else {
          gap = false;
        }
      }
    } else {
      gap = true;
    }
  }

  return ret;
}

/// Validate data array.
bool validateData(List data, Map<int, EdiTeXTemplate> templates) {
  final argiStack = new List<int>();
  final argcStack = new List<int>();

  for (final v in data) {
    if (v is int) {
      if (v >= 0) {
        // At this point is is possible that no template exists for this ID,
        // which means we return false.
        if (!templates.containsKey(v) || templates[v] == null) {
          return false;
        }

        final parameterCount = templates[v].parameterCount;
        if (parameterCount > 0) {
          argiStack.add(1);
          argcStack.add(parameterCount);
        }
      } else {
        // If stack is empty, return false.
        if (argiStack.isEmpty) {
          return false;
        }

        // Check that this boundary is as expected.
        if (argiStack.last != -v) {
          return false;
        }

        // Pop stack if this is the last argument.
        if (argiStack.last == argcStack.last) {
          argiStack.removeLast();
          argcStack.removeLast();
        } else {
          // Increment to next parameter.
          argiStack[argiStack.length - 1]++;
        }
      }
    }
  }

  // Both stacks should be empty here.
  return argiStack.isEmpty;
}
